import {ScriptAccountCondition} from "../conditions/ScriptAccountCondition";
import {ScriptAccountAction} from "../actions/ScriptAccountAction";
import {ProductState} from "../states/ProductState";
import {State} from "../states/State";
import {Transition} from "../transitions/Transition";
import {ProductGamesystem} from "../ProductGamesystem";
import {ProductGeneratorResult} from "./ProductGeneratorResult";
import {Gamesystem} from "../Gamesystem";
import {ProductGenerationData} from "./ProductGenerationData";
import {ProductTransition} from "../transitions/ProductTransition";

export class ProductSystemGenerator {
  productGamesystem: ProductGamesystem


  constructor(productGamesystem: ProductGamesystem) {
    this.productGamesystem = productGamesystem;
  }

  generateFromChildsystems(): void {
    if(this.productGamesystem.innerGamesystems.length < 2) return;
    const leftInitialData = this.prepareChildsystemForGeneration(this.productGamesystem.innerGamesystems[0])
    const rightInitialData = this.prepareChildsystemForGeneration(this.productGamesystem.innerGamesystems[1])

    const initialGenerationResult = this.generateFromBinaryChildsystems(leftInitialData, rightInitialData);

    if(this.productGamesystem.innerGamesystems.length > 2) {
      for(let i=2; i<this.productGamesystem.innerGamesystems.length; i++) {
        const leftData = initialGenerationResult.productGenerationData;
        const rightData = this.prepareChildsystemForGeneration(this.productGamesystem.innerGamesystems[i]);

        const generationResult = this.generateFromBinaryChildsystems(leftData, rightData);
        this.assignGeneratedStatesAndTransitions(generationResult)
      }
    } else {
      this.assignGeneratedStatesAndTransitions(initialGenerationResult)
    }
  }

  protected prepareChildsystemForGeneration(childsystem: Gamesystem<any, any>) {
    if(childsystem instanceof ProductGamesystem) {
      childsystem.generateFromChildsystems()
    }

    return new ProductGenerationData(childsystem.states, childsystem.transitions)
  }

  protected generateFromBinaryChildsystems(leftSystemData: ProductGenerationData, rightSystemData: ProductGenerationData): ProductGeneratorResult {
    const generatedProductStates: ProductState[] = []
    const generatedProductTransitions: ProductTransition[] = []

    leftSystemData.states.forEach(leftState => {
      rightSystemData.states.forEach(rightState => {
        for(let i=0; i<leftState.outgoingTransitions.length; i++) {
          for(let j=0; j<rightState.outgoingTransitions.length; j++) {
            const startingState = this.generateBinaryProductState(leftState, rightState, generatedProductStates);

            if(startingState != undefined) {
              const endingState_right = this.generateBinaryProductState(leftState, rightState.outgoingTransitions[j].endingState, generatedProductStates);
              if(endingState_right != undefined) {
                this.generateBinaryProductTransition(startingState, endingState_right, rightState.outgoingTransitions[j], generatedProductTransitions);
              }

              const endingState_left = this.generateBinaryProductState(leftState.outgoingTransitions[i].endingState, rightState, generatedProductStates);
              if(endingState_left != undefined) {
                this.generateBinaryProductTransition(startingState, endingState_left, leftState.outgoingTransitions[i], generatedProductTransitions)
              }

              const endingState_left_right = this.generateBinaryProductState(leftState.outgoingTransitions[i].endingState, rightState.outgoingTransitions[j].endingState, generatedProductStates);
              if(endingState_left_right != undefined) {
                this.generateBinaryProductTransitionMulti(startingState, endingState_left_right, leftState.outgoingTransitions[i], rightState.outgoingTransitions[j], generatedProductTransitions);
              }
            }
          }

          if(rightState.outgoingTransitions.length == 0) {
            const startingState = this.generateBinaryProductState(leftState, rightState, generatedProductStates);
            const endingState = this.generateBinaryProductState(leftState.outgoingTransitions[i].endingState, rightState, generatedProductStates);

            if(startingState != undefined && endingState != undefined) {
              this.generateBinaryProductTransition(startingState, endingState, leftState.outgoingTransitions[i], generatedProductTransitions)

            }
          }
        }

        if(leftState.outgoingTransitions.length == 0) {
          for(let j=0; j<rightState.outgoingTransitions.length; j++) {
            const startingState = this.generateBinaryProductState(leftState, rightState, generatedProductStates);
            const endingState = this.generateBinaryProductState(leftState, rightState.outgoingTransitions[j].endingState, generatedProductStates);

            if(startingState != undefined && endingState != undefined) {
              this.generateBinaryProductTransition(startingState, endingState, rightState.outgoingTransitions[j], generatedProductTransitions);

            }

          }
        }
      })
    })

    return new ProductGeneratorResult(generatedProductStates, generatedProductTransitions);
  }

  protected assignGeneratedStatesAndTransitions(generationResult: ProductGeneratorResult): void {
    this.productGamesystem.states = generationResult.states;
    this.productGamesystem.transitions = generationResult.transitions;
  }

  protected generateBinaryProductTransition(startingState: ProductState, endingState: ProductState, usedTransition: Transition<any>, generatedTransitions: ProductTransition[]) {
    const transition = new ProductTransition(startingState, endingState);
    transition.scriptAccountActions = [... usedTransition.scriptAccountActions];
    transition.scriptAccountConditions = [... usedTransition.scriptAccountConditions];

    if(generatedTransitions.find(generatedTransition => generatedTransition.startingState.equals(startingState) && generatedTransition.endingState.equals(endingState)) == undefined) {
      generatedTransitions.push(transition)
    }
  }

  protected generateBinaryProductTransitionMulti(startingState: ProductState, endingState: ProductState, leftTransition: Transition<any>, rightTransition: Transition<any>, generatedTransitions: ProductTransition[]) {
    const leftConditions = leftTransition.scriptAccountConditions;
    const rightConditions = rightTransition.scriptAccountConditions;

    if(!this.contradictCombinedConditions(leftConditions, rightConditions)) {
      const transition = new ProductTransition(startingState, endingState)
      transition.scriptAccountActions = this.generateCombinedActions(leftTransition.scriptAccountActions, rightTransition.scriptAccountActions);
      transition.scriptAccountConditions = this.generateCombinedConditions(leftTransition.scriptAccountConditions, rightTransition.scriptAccountConditions);

      if(generatedTransitions.find(generatedTransition => generatedTransition.startingState.equals(startingState) && generatedTransition.endingState.equals(endingState)) == undefined) {
        generatedTransitions.push(transition)
      }
    }
  }

  protected generateBinaryProductState(leftState: State<any>, rightState: State<any>, generatedStates: ProductState[]): ProductState | undefined {
    const combinedStateConditions: ScriptAccountCondition[] = leftState.conditions.concat(rightState.conditions);
    for(let i=0; i<combinedStateConditions.length; i++) {
      for(let j=0; j<combinedStateConditions.length; j++) {
        if(combinedStateConditions[i].isContradicting(combinedStateConditions[j])) {
          return undefined
        }
      }
    }


    let innerStates: State<any>[] = []
    if(leftState instanceof ProductState) {
      innerStates = leftState.innerStates
    } else {
      innerStates = [leftState]
    }

    if(rightState instanceof ProductState) {
      innerStates = innerStates.concat(rightState.innerStates)
    } else {
      innerStates = innerStates.concat(rightState)
    }

    let binaryProductState: ProductState | undefined = this.findGeneratedProductState(innerStates, generatedStates);
    if(binaryProductState == undefined) {
      binaryProductState = new ProductState(innerStates);
      generatedStates.push(binaryProductState)
    }

    //binaryProductState.determineInitialProperty()
    binaryProductState.conditions = combinedStateConditions;

    return binaryProductState;
  }


  protected findGeneratedProductState(innerStates: State<any>[], productStates: ProductState[]) {
    return productStates.find(productState => productState.equalInnerStates(innerStates));
  }

  protected generateCombinedActions(leftActions: ScriptAccountAction[], rightActions: ScriptAccountAction[]): ScriptAccountAction[] {
    const combinedActions: ScriptAccountAction[] = []
    for(let i=0; i<leftActions.length; i++) {
      for(let j=0; j<rightActions.length; j++) {
        const combinedAction = leftActions[i].combineActions(rightActions[j])
        if(combinedAction == undefined) {
          if(!combinedActions.includes(leftActions[i])) {
            combinedActions.push(leftActions[i])
          }

          if(!combinedActions.includes(rightActions[j])) {
            combinedActions.push(rightActions[j])
          }
        } else {

          combinedActions.push(combinedAction)
        }
      }
    }
    return combinedActions;
  }

  protected generateCombinedConditions(leftConditions: ScriptAccountCondition[], rightConditions: ScriptAccountCondition[]): ScriptAccountCondition[] {
    const combinedConditions: ScriptAccountCondition[] = [];
    for(let i=0; i<leftConditions.length; i++) {
      for(let j=0; j<rightConditions.length; j++) {
        const combinedCondition = leftConditions[i].combineCondition(rightConditions[j]);
        if(combinedCondition == undefined) {
          if(!combinedConditions.includes(leftConditions[i])) {
            combinedConditions.push(leftConditions[i])
          }

          if(!combinedConditions.includes(rightConditions[j])) {
            combinedConditions.push(rightConditions[j])
          }
        } else {
          combinedConditions.push(combinedCondition)
        }
      }
    }
    return combinedConditions;
  }

  protected contradictCombinedConditions(leftConditions: ScriptAccountCondition[], rightConditions: ScriptAccountCondition[]): boolean {
    for(let i=0; i<leftConditions.length; i++) {
      for(let j=0; j<rightConditions.length; j++) {
        if(leftConditions[i].isContradicting(rightConditions[j])) {
          return true;
        }
      }
    }
    return false;
  }

}
